"""Contains setting parameters for the library. """

PIP_PROJECT_NAME = 'lazycluster'
GITHUB_URL = 'https://github.com/ml-tooling/lazycluster.git'
BRANCH = 'develop'  # Leave empty to enforce lazycluster installation from PyPi - see _utils.get_pip_install_cmd()